from typing import Optional, List, Dict
import logging
import asyncio
import websockets
from rpc_gateway import messages, errors
from rpc_gateway.message_pump import MessagePump

logger = logging.getLogger(__name__)


class GatewayConnection:
    def __init__(self, gateway: 'Gateway', connection: websockets.WebSocketServerProtocol):
        self.gateway = gateway
        self.connection = connection
        self.message_pump = MessagePump(connection, lambda message: gateway.handle_request(self, message))
        self.instances: List[str] = []

    async def start(self):
        await self.message_pump.start()

    async def stop(self):
        await self.message_pump.stop()

    def register_instances(self, instances):
        self.instances = list(set(*self.instances, *instances))


class Gateway:
    SERVER_MESSAGES = (messages.Method.GET, messages.Method.SET, messages.Method.CALL)

    def __init__(self, host: str = 'localhost', port: int = 8888):
        self.host = host
        self.port = port
        self.logger = logger.getChild(self.__class__.__name__)
        self.gateway_connections: Dict[websockets.WebSocketCommonProtocol, GatewayConnection] = {}
        self.websocket: Optional[websockets.WebSocketServer] = None
        self.websocket_task: Optional[asyncio.Task] = None
        self.event_loop = asyncio.get_event_loop()
        self.instances: Dict[str, GatewayConnection] = {}

    def start(self, wait = True):
        self.event_loop.run_until_complete(self._start(wait))

    async def _start(self, wait = True):
        self.logger.info(f'Starting on ws://{self.host}:{self.port}')
        self.websocket = await websockets.serve(self.on_connection, self.host, self.port)
        self.websocket_task = asyncio.Task(self.websocket.wait_closed())

        if wait:
            await self._wait()

    def wait(self):
        self.event_loop.run_until_complete(self._wait())

    async def _wait(self):
        if self.websocket_task is not None:
            await self.websocket_task

        self.logger.info(f'Done')

    def stop(self):
        self.event_loop.run_until_complete(self._stop())

    async def _stop(self):
        await asyncio.gather(*[gateway_connection.stop() for w, gateway_connection in self.gateway_connections.items()])
        self.websocket_task.cancel()

    async def on_connection(self, connection: websockets.WebSocketServerProtocol, path: str):
        gateway_connection = self.gateway_connections[connection] = GatewayConnection(self, connection)
        await gateway_connection.start()
        self.gateway_connections.pop(connection)

    #
    # Request Handlers
    #

    async def handle_forward_request(self, gateway_connection: GatewayConnection, request: messages.Request) -> messages.Response:
        self.logger.debug(f'Forwarding request to server: {request}')
        server = self.instances[request.data['instance']]
        response = await server.message_pump.request(request, raise_error=False)
        self.logger.debug(f'Forwarding response to client: {response}')

        return response

    async def handle_available_request(self, gateway_connection: GatewayConnection, request: messages.Request) -> messages.Response:
        available = request.data['instance'] in self.instances
        return messages.Response(data=available)

    async def handle_register_request(self, gateway_connection: GatewayConnection, request: messages.Request) -> messages.Response:
        self.logger.debug(f'Registering instances: {request.data}')
        gateway_connection.register_instances(request.data)
        for instance_name in request.data:
            self.instances[instance_name] = gateway_connection

        return messages.Response()

    # this is called by the GatewayConnection MessagePump when a new request is received
    async def handle_request(self, gateway_connection: GatewayConnection, request: messages.Request) -> messages.Response:
        if request.method in self.SERVER_MESSAGES:
            return await self.handle_forward_request(gateway_connection, request)

        if request.method == messages.Method.AVAILABLE:
            return await self.handle_available_request(gateway_connection, request)

        if request.method == messages.Method.REGISTER:
            return await self.handle_register_request(gateway_connection, request)

        return messages.Response(id=request.id, data=errors.InvalidMethodError(f'Invalid method: {request.method}'))


class GatewayClient:
    def __init__(self, gateway_url: str = 'ws://localhost:8888'):
        self.logger = logger.getChild(self.__class__.__name__)
        self.gateway_url = gateway_url
        self.message_pump = MessagePump(request_handler=self._handle_request)
        self.event_loop = asyncio.get_event_loop()

    @property
    def connected(self) -> bool:
        return self.message_pump.connection is not None

    def start(self, wait=True):
        self.event_loop.run_until_complete(self._start(wait))

    async def _start(self, wait=True):
        self.connection = await websockets.connect(self.gateway_url)
        await self.message_pump.start(wait=False, connection=self.connection)
        await self._on_start()

        if wait:
            await self._wait()

    def wait(self):
        self.event_loop.run_until_complete(self._wait())

    async def _wait(self):
        await self.message_pump.wait()

    async def _on_start(self):
        pass

    def stop(self):
        self.event_loop.run_until_complete(self._stop())

    async def _stop(self):
        await self.message_pump.stop()

    async def _handle_request(self, request: messages.Request) -> messages.Response:
        pass


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)

    gateway = Gateway()
    gateway.start()


from typing import Optional, Dict, Any, Callable, Awaitable
import asyncio
import pickle
import logging
import websockets
from rpc_gateway import messages, errors

logger = logging.getLogger(__name__)
RequestHandlerType = Callable[[messages.Request], Awaitable[messages.Response]]
MAX_MESSAGE_ID = 99999
last_message_id = 0


def next_message_id() -> int:
    global last_message_id
    last_message_id += 1
    return last_message_id % MAX_MESSAGE_ID


class MessagePump:
    next_id = 0

    def __init__(self,
                 connection: Optional[websockets.WebSocketCommonProtocol] = None,
                 request_handler: Optional[RequestHandlerType] = None,
                 message_queues: Optional[Dict[int, asyncio.Queue]] = None):
        self.id = MessagePump.next_id
        MessagePump.next_id += 1
        self.logger = logger.getChild(self.__class__.__name__)
        self.connection = connection
        self.request_handler = request_handler
        self.event_loop = asyncio.get_event_loop()
        self.message_pump_task: Optional[asyncio.Task] = None
        self.message_queues = {} if message_queues is None else message_queues
        self.running = False

    async def start(self, wait=True, connection: Optional[websockets.WebSocketCommonProtocol] = None):
        if connection is not None:
            self.connection = connection

        self.running = True
        self.message_pump_task = asyncio.Task(self.message_pump())

        if wait:
            await self.wait()

    async def wait(self):
        await self.message_pump_task

    async def stop(self):
        self.running = False
        await self.connection.close_connection()
        self.message_pump_task.cancel()

    async def send_message(self, message: messages.Message):
        if self.connection is None:
            raise errors.NotConnectedError('Must be connected to send message: {message}')

        self.logger.debug(f'[{self.id}] > {message}')
        await self.connection.send(message.to_pickle())

    async def send_request(self, method: str, data: Any = None):
        return await self.request(messages.Request(method=method, data=data))

    async def send_response(self, id: int, data: Any = None, status: int = messages.Status.SUCCESS):
        response = messages.Response(id=id, status=status, data=data)
        await self.send_message(response)

    async def send_error_response(self, id: int, error: Exception):
        response = messages.Response(id=id, status=messages.Status.ERROR, data=error)
        await self.send_message(response)

    async def request(self, request: messages.Request, raise_error=True) -> messages.Response:
        request.id = next_message_id()
        queue = self.message_queues[request.id] = asyncio.Queue()
        await self.send_message(request)
        response: messages.Response = await queue.get()
        self.message_queues.pop(request.id)

        if response.status == messages.Status.ERROR and raise_error:
            raise response.data

        return response

    async def handle_request(self, request: messages.Request):
        response = await self.request_handler(request)
        response.id = request.id
        await self.send_message(response)

    async def message_pump(self):
        while self.running:
            message_bytes = await self.connection.recv()
            message = pickle.loads(message_bytes)
            self.logger.debug(f'[{self.id}] < {message}')

            if isinstance(message, messages.Response):  # response
                if message.id not in self.message_queues:
                    raise errors.InvalidMessageIdError(f'No request found for response ID: {message.id}')

                queue = self.message_queues[message.id]
                await queue.put(message)

            elif isinstance(message, messages.Request):  # request
                if self.request_handler is not None:
                    asyncio.create_task(self.handle_request(message))

            else:
                raise errors.InvalidMessageError(f'Invalid message: {message}')
